// *** GAME CLIENT CODE *** \\

// sounds!
snd_cannon = new Audio("../public/sounds/cannon.wav");
snd_small_explosion = new Audio("../public/sounds/small_explosion.wav");
snd_med_explosion = new Audio("../public/sounds/med_explosion.wav");
snd_large_explosion = new Audio("../public/sounds/large_explosion.wav");
snd_start_level = new Audio("../public/sounds/start_level.wav");
snd_player_explosion = new Audio("../public/sounds/player_explosion.wav");
snd_oh_man = new Audio("../public/sounds/oh_man.wav");

/**************************************************
** GAME VARIABLES
**************************************************/
var canvas,			// Canvas DOM element
	ctx,			// Canvas rendering context
	keys,			// Keyboard input
	localPlayer,	// Local player
	remotePlayers,	// remote players
	projectiles,	// weapons (projectiles)
	asteroids,		// asteroids
	socket;			// Socket connection

/**************************************************
** GAME INITIALISATION
**************************************************/
function init() {
	// Declare the canvas and rendering context
	canvas = document.getElementById("gameCanvas");
	ctx = canvas.getContext("2d");

	// Maximise the canvas
	canvas.width = window.innerWidth;
	canvas.height = window.innerHeight;

	// Initialise keyboard controls
	keys = new Keys();

	// Calculate a random start position for the local player
	// The minus 5 (half a player size) stops the player being
	// placed right on the egde of the screen
	var startX = Math.round(Math.random()*(canvas.width-5)),
		startY = Math.round(Math.random()*(canvas.height-5));

	// Initialise the local player
	localPlayer = new Player(startX, startY);
	
	// connect to a Socket.IO server.
	// the first parameter specifies the server address (localhost for now).
	// if we want to point this at another server for a demo, need to replace 'localhost' with the 
	// external IP of the server:
	socket = io.connect('http://localhost', {port: 8000, transports: ["websocket"]}); 
		// i.e. 169.234.24.134 vice localhost to connect to a Node instance running somewhere else
	
	// initialize the remote players
	remotePlayers = [];
	
	// initialize the weapons (projectiles)
	projectiles = [];
	
	// initialize the asteroids
	asteroids = [];
	
	// Start listening for events
	setEventHandlers();
};

/**************************************************
** GAME EVENT HANDLERS
**************************************************/
var setEventHandlers = function() {
	// Keyboard
	window.addEventListener("keydown", onKeydown, false);
	window.addEventListener("keyup", onKeyup, false);
	//window.addEventListener("keypress", onKeypress, false); - not currently being used; was messing around with using it to create different firing modes

	// Window resize
	window.addEventListener("resize", onResize, false);
	
	// socket connection successful
	socket.on("connect", onSocketConnected);
	// socket disconnection
	socket.on("disconnect", onSocketDisconnect);
	// new player message received
	socket.on("new player", onNewPlayer);
	// player move message received
	socket.on("move player", onMovePlayer);
	// player removed message received
	socket.on("remove player", onRemovePlayer);
	// player fired message received
	socket.on("fire player", onFirePlayer);
	// new asteroid created message received
	socket.on("new asteroid", onNewAsteroid);
	// remove asteroid message received
	socket.on("remove asteroid", onRemoveAsteroid);
	// remove projectile message received
	socket.on("remove projectile", onRemoveProjectile);
	// you're dead message received - killed by an asteroid
	socket.on("you're dead", onYoureDead);
	// you're dead friendly fire message received - killed by an asteroid
	socket.on("you're dead friendly fire", onYoureDeadFriendlyFire);
};

// Keyboard key down
function onKeydown(e) {
	if (localPlayer) {
		keys.onKeyDown(e);
	};
};

// Keyboard key up
function onKeyup(e) {
	if (localPlayer) {
		keys.onKeyUp(e);
	};
};

// Keyboard key press
function onKeypress(e) {
	if (localPlayer) {
		keys.onKeypress(e);
	};
};

// Browser window resize
function onResize(e) {
	// Maximise the canvas
	canvas.width = window.innerWidth;
	canvas.height = window.innerHeight;
};

// called when a socket connection is made
function onSocketConnected() {
	console.log("Connected to socket server");
	snd_start_level.play();
	// tell the server to create a new player when successfully connected:
	socket.emit("new player", {x: localPlayer.getX(), y: localPlayer.getY()});
	// start the asteroid loop - tell the server to create a new asteroid every so often.
	// this way, the more connected players there are, the more asteroids are created.
	// also, this avoids the complexity of having server-initiated events;
	// instead, the server just responds to each of these requests by creating a new asteroid,
	// and notifying all connected clients.
	startAsteroidLoop();
};

// tell the server to create a new asteroid every so often.
// the server will create the asteroid and reply to all connected clients with a
// "new asteroid" event.
function startAsteroidLoop() {
	setInterval(function() {
		socket.emit("create asteroid");
	}, 3000);
};

// called when a socket connection is broken
function onSocketDisconnect() {
	console.log("Disconnected from socket server");
};

// called when a new player is connected
function onNewPlayer(data) {
	console.log("New player connected: "+data.id);
	// create a new player object based on position data sent from the server.
	var newPlayer = new Player(data.x, data.y);
	newPlayer.id = data.id;
	// add the new Player object to the remotePlayers array so we can access it later.
	remotePlayers.push(newPlayer);
};

// called when a player moves
function onMovePlayer(data) {
	// search for the player that is being moved 
	var movePlayer = playerById(data.id);

	if (!movePlayer) {
		console.log("Player not found: "+data.id);
		return;
	};

	// update the player's x and y position
	movePlayer.setX(data.x);
	movePlayer.setY(data.y);
};

// called when a player fires a weapon
function onFirePlayer(data) {
	var newProjectile = new Projectile(data.x, data.y);
	
	// add the new projectile to the projectiles array
	projectiles.push(newProjectile);
};

// called when a player is removed
function onRemovePlayer(data) {
	var removePlayer = playerById(data.id);
	
	// find the player to remove based on his id number
	if (!removePlayer) {
		console.log("Player not found: "+data.id);
		return;
	};
	
	// remove the disconnected player from the remotePlayers array	
	remotePlayers.splice(remotePlayers.indexOf(removePlayer), 1);
};

// handle the asteroid loop
function onNewAsteroid(data) {
	var aNum = Math.floor(Math.random()*5);
	var newAsteroid = new Asteroid(data.x, data.y, data.xSpeed, data.ySpeed, data.aNum);
	
	// add the new asteroid to the asteroids array
	asteroids.push(newAsteroid);
};

function onRemoveAsteroid(data) {
	// remove the destroyed asteroid from the asteroids array
	asteroids.splice(data.index, 1);
};

function onRemoveProjectile(data) {
	// remove the destroyed projectile from the projectiles array
	projectiles.splice(data.index, 1);
};

// alert message to be displayed when the player is killed by an asteroid.
// acknowledging reloads the page, and you're back in the game.
function onYoureDead() {
	alert("You've been killed by an asteroid!\n" + "Refresh the page to rejoin the fight!");
	location.reload(true);
};

// alert message to be displayed when the player is killed by friendly fire (projectiles).
// acknowledging reloads the page, and you're back in the game.
function onYoureDeadFriendlyFire() {
	alert("You've been killed by friendly fire!\n" + "Refresh the page to rejoin the fight!");
	location.reload(true);
};

// helper function to locate player within the players array, using supplied id:
function playerById(id) {
	var i;
	for (i = 0; i < remotePlayers.length; i++) {
		if (remotePlayers[i].id == id)
			return remotePlayers[i];
	};
	
	return false;
};

/**************************************************
** GAME ANIMATION LOOP
**************************************************/
function animate() {
	update();
	draw();

	// Request a new animation frame using Paul Irish's shim
	window.requestAnimFrame(animate);
};

/**************************************************
** GAME UPDATE
**************************************************/
function update() {
	// send the player position to the server after every update,
	// only if the player position has changed
	if (localPlayer.update(keys)) {
		socket.emit("move player", {x: localPlayer.getX(), y: localPlayer.getY()});
	};
	// send the position of a newly fired projectile, 
	// only if the player has fired
	if (localPlayer.fire(keys)) {
		socket.emit("fire player", {x: localPlayer.getX(), y: localPlayer.getY() - 25}); // shoot from the front of the ship, not the center
		snd_cannon.play();
		// we'd like the cannon sound to fire on each button press, so you can create a "rapid-fire"
		// effect, so try and restart the sound.
		// not all browsers handle this correctly though, so wrap it in a try / catch to be safe.
		try {
			snd_cannon.currentTime = 0;
		} catch (err) {
			console.log("error playing sound: "+err);
		}
	};

	// move the projectiles:
	for (i = 0; i < projectiles.length; i++) {
		projectiles[i].update();
		// delete any projectiles that have travelled off screen:
		if (projectiles[i].getY() < 0) {
			socket.emit("projectile off screen", {projectileIndex: i});
		};
	};
	
	// move the asteroids:
	for (i = 0; i < asteroids.length; i++) {
		asteroids[i].update();
		// delete any asteroids that have travelled off screen:
		// note that the size of the canvas varies with the player's window size, so just checking
		// against canvas.height might mean that we destroy asteroids that one player can't see 
		// anymore, even if players with larger resolutions might still be able to see.
		// because of this, we just check against a Y value > 1000 as a conservative measure.
		if (asteroids[i].getY() > 1000) {
			socket.emit("asteroid off screen", {asteroidIndex: i});
		};
	};
	
	// *** collision detection ***
	
	// look for collisions between asteroids and projectiles
	var i, j; // loop counters
	var centerAsteroidX, centerAsteroidY, centerProjectileX, centerProjectileY; // center points of the asteroid and projectile bounding circles
	var radiusAsteroid, radiusProjectile; 										// radii of the asteroid and projectile bounding circles
	var distX, distY;
	var squareDist;
	radiusProjectile = 3.25; // based on the 13x12px bullet image
	for (i = 0; i < asteroids.length; i++) {
		for (j = 0; j < projectiles.length; j++) {
			centerAsteroidX = asteroids[i].getX();
			centerAsteroidY = asteroids[i].getY();
			centerProjectileX = projectiles[j].getX();
			centerProjectileY = projectiles[j].getY();
			// figure out what the asteroid's radius is for the bounding sphere.
			// this is based on the size of images a0 - a4.
			switch (asteroids[i].get_aNum()) {
				case 0:
					radiusAsteroid = 5;
					break;
				case 1:
					radiusAsteroid = 10;
					break;
				case 2: 
					radiusAsteroid = 15;
					break;
				case 3:
					radiusAsteroid = 20;
					break;
				case 4: 
					radiusAsteroid = 25;
					break;
			}
			// calculate distance between centers
			distX = centerAsteroidX - centerProjectileX;
			distY = centerAsteroidY - centerProjectileY;
			// get squared distance and compare to square of the sum of the radii
			squareDist = ((distX * distX) + (distY * distY));
			// detect collision
			if (squareDist <= ((radiusAsteroid + radiusProjectile) * (radiusAsteroid * radiusProjectile))) {
				socket.emit("projectile collision", {asteroidIndex: i, projectileIndex: j});
				if (asteroids[i].get_aNum() == 0 || asteroids[i].get_aNum() == 1) {
					snd_small_explosion.play();
				} else if (asteroids[i].get_aNum() == 2 || asteroids[i].get_aNum() == 3) {
					snd_med_explosion.play();
				} else {
					snd_large_explosion.play();
				}
			};
		};
	};
	
	// look for collisions between asteroids and the local player
	var i; // loop counter
	var centerAsteroidX, centerAsteroidY, centerplayerX, centerPlayerY; // center points of the asteroid and player bounding circles
	var radiusAsteroid, radiusPlayer;									// radii of the asteroid and player bounding circles
	var distX, distY;
	var squareDist;
	radiusPlayer = 6.25; // based on the 25x49px spaceship images
	for (i = 0; i < asteroids.length; i++) {
		centerAsteroidX = asteroids[i].getX();
		centerAsteroidY = asteroids[i].getY();
		centerPlayerX = localPlayer.getX();
		centerPlayerY = localPlayer.getY();
		// figure out what the asteroid's radius is for the bounding sphere.
		// this is based on the size of images a0 - a4.
		switch (asteroids[i].get_aNum()) {
			case 0:
				radiusAsteroid = 5;
				break;
			case 1:
				radiusAsteroid = 10;
				break;
			case 2: 
				radiusAsteroid = 15;
				break;
			case 3:
				radiusAsteroid = 20;
				break;
			case 4: 
				radiusAsteroid = 25;
				break;
		}
		// calculate distance between centers
		distX = centerAsteroidX - centerPlayerX;
		distY = centerAsteroidY - centerPlayerY;
		// get squared distance and compare to square of the sum of the radii
		squareDist = ((distX * distX) + (distY * distY));
		// detect collision
		if (squareDist <= ((radiusAsteroid + radiusPlayer) * (radiusAsteroid * radiusPlayer))) {
			socket.emit("player collision");
			snd_player_explosion.play();
		};		
	};
	
	// look for collisions between projectiles and the local player
	// THIS IS DEATH BY FRIENDLY FIRE!!!
	var i; // loop counter
	var centerProjectileX, centerProjectileY, centerplayerX, centerPlayerY; // center points of the projectile and player bounding circles
	var radiusAsteroid, radiusPlayer;										// radii of the asteroid and player bounding circles
	var distX, distY;
	var squareDist;
	radiusProjectile = 3.25; // based on the 13x12px bullet image
	radiusPlayer = 6.25; 	// based on the 25x49px spaceship images
	for (i = 0; i < projectiles.length; i++) {
		centerProjectileX = projectiles[i].getX();
		centerProjectileY = projectiles[i].getY() - 15; // this small offset prevents players from getting insta-killed by their own projectiles
		centerPlayerX = localPlayer.getX();
		centerPlayerY = localPlayer.getY();
		// calculate distance between centers
		distX = centerProjectileX - centerPlayerX;
		distY = centerProjectileY - centerPlayerY;
		// get squared distance and compare to square of the sum of the radii
		squareDist = ((distX * distX) + (distY * distY));
		// detect collision
		if (squareDist <= ((radiusProjectile + radiusPlayer) * (radiusProjectile * radiusPlayer))) {
			socket.emit("friendly fire collision", {projectileIndex: i});
			snd_oh_man.play();
		};		
	};
};

/**************************************************
** GAME DRAW
**************************************************/
function draw() {
	// Wipe the canvas clean
	ctx.clearRect(0, 0, canvas.width, canvas.height);

	// Draw the local player
	localPlayer.drawSelf(ctx);
	
	// loop through every remote player, and draw it to the canvas.
	var i;
	for (i = 0; i < remotePlayers.length; i++) {
		remotePlayers[i].draw(ctx);
	};
	
	// loop through every projectile, and draw it to the canvas.
	for (i = 0; i < projectiles.length; i++) {
		projectiles[i].draw(ctx);
	};
	
	// loop through every asteroid, and draw it to the canvas.
	for (i = 0; i < asteroids.length; i++) {
		asteroids[i].draw(ctx);
	};
};

// *** =============================================================================================
// *** (commented out, not-fully-working bits below)

	// this is commented out b/c it wasn't working -- 
	// but it doesn't really matter.
	// this was only something I was trying b/c if you open multiple tabs from the same browser,
	// players killed on "unattended" tabs don't tell the server they are dead, so it can create
	// some weirdness.
	// i'm OK with just living with that for now.
	/*
	// look for collisions between asteroids and the remote players
	var i, j; // loop counters
	var centerAsteroidX, centerAsteroidY, centerPlayerX, centerPlayerY;	// center points of the asteroid and player bounding circles
	var radiusAsteroid, radiusPlayer;									// radii of the asteroid and player bounding circles
	var distX, distY;
	var squareDist;
	radiusPlayer = 6.25; // based on the 25x49px spaceship images
	for (i = 0; i < asteroids.length; i++) {
		for (j = 0; j < remotePlayers.length; j++) {
			centerAsteroidX = asteroids[i].getX();
			centerAsteroidY = asteroids[i].getY();
			centerPlayerX = remotePlayers[j].getX();
			centerPlayerY = remotePlayers[j].getY();
			// figure out what the asteroid's radius is for the bounding sphere.
			// this is based on the size of images a0 - a4.
			switch (asteroids[i].get_aNum()) {
				case 0:
					radiusAsteroid = 5;
					break;
				case 1:
					radiusAsteroid = 10;
					break;
				case 2: 
					radiusAsteroid = 15;
					break;
				case 3:
					radiusAsteroid = 20;
					break;
				case 4: 
					radiusAsteroid = 25;
					break;
			}
			// calculate distance between centers
			distX = centerAsteroidX - centerPlayerX;
			distY = centerAsteroidY - centerPlayerY;
			// get squared distance and compare to square of the sum of the radii
			squareDist = ((distX * distX) + (distY * distY));			
			// detect collision
			if (squareDist <= ((radiusAsteroid + radiusPlayer) * (radiusAsteroid * radiusPlayer))) {
				console.log("remote player collision detected!");
				console.log("remote player id is: "+remotePlayers[j].id);
				// socket needs to emit an event here, and the id of the player who has been killed needs to get passed to the server
				socket.emit("remote player collision", {deadPlayerID: remotePlayers[j].id});
				snd_player_explosion.play();
			};		
		};
	};
	*/
